import tarfile
import urllib.error
import urllib.request
from pathlib import Path
from time import sleep

from .__version import __version__

HOME_BRIDGESTAN = Path("~").expanduser() / ".bridgestan"
CURRENT_BRIDGESTAN = HOME_BRIDGESTAN / f"bridgestan-{__version__}"

RETRIES = 5


def get_bridgestan_src():
    """
    Download and unzip the BridgeStan source distribution for this version

    Based on similar code from cmdstanpy's install_cmdstan script
    """
    url = (
        "https://github.com/roualdes/bridgestan/releases/download/"
        + f"v{__version__}/bridgestan-{__version__}.tar.gz"
    )
    HOME_BRIDGESTAN.mkdir(exist_ok=True)

    err_text = f"Failed to download Bridgestan {__version__} from github.com."
    for i in range(1, 1 + RETRIES):
        try:
            file_tmp, _ = urllib.request.urlretrieve(url, filename=None)
            break
        except urllib.error.HTTPError as e:
            # not recoverable
            raise ValueError(err_text) from e
        except urllib.error.URLError as e:
            print(err_text)
            if i == RETRIES:
                raise ValueError(err_text) from e
            else:
                print("Retrying ({i+1}/{RETRIES})...")
                sleep(1)

    try:
        tar = tarfile.open(file_tmp)
        tar.extractall(path=HOME_BRIDGESTAN)
    except Exception as e:
        raise ValueError(f"Failed to unpack {file_tmp} during installation") from e
    finally:
        tar.close()



import os
import platform
import subprocess
from pathlib import Path
from typing import List

from .__version import __version__
from .download import CURRENT_BRIDGESTAN, get_bridgestan_src
from .util import validate_readable


def verify_bridgestan_path(path: str) -> None:
    folder = Path(path).resolve()
    if not folder.exists():
        raise ValueError(
            f"BridgeStan folder '{folder}' does not exist!\n"
            "If you need to set a different location, call 'set_bridgestan_path()'"
        )
    makefile = folder / "Makefile"
    if not makefile.exists():
        raise ValueError(
            f"BridgeStan folder '{folder}' does not "
            "contain file 'Makefile', please ensure it is built properly!\n"
            "If you need to set a different location, call 'set_bridgestan_path()'"
        )


PYTHON_FOLDER = Path(__file__).parent.parent

MAKE = os.getenv(
    "MAKE",
    "make" if platform.system() != "Windows" else "mingw32-make",
)


def set_bridgestan_path(path: str) -> None:
    """Set the path to BridgeStan.

    This should point to the top-level folder of the repository.

    By default this is set to the value of the environment
    variable ``BRIDGESTAN``, or to the folder above the location
    of this package (which, assuming a source installation, corresponds
    to the repository root).
    """
    verify_bridgestan_path(path)
    os.environ["BRIDGESTAN"] = path


def get_bridgestan_path():
    path = os.getenv("BRIDGESTAN", "")
    if path == "":
        try:
            path = str(CURRENT_BRIDGESTAN)
            verify_bridgestan_path(path)
        except ValueError:
            print(
                "Bridgestan not found at location specified by $BRIDGESTAN "
                f"environment variable, downloading version {__version__} to {path}"
            )
            get_bridgestan_src()
            print("Done!")

    return path


def generate_so_name(model: Path):
    name = model.stem
    return model.with_stem(f"{name}_model").with_suffix(".so")


def compile_model(
    stan_file: str, *, stanc_args: List[str] = [], make_args: List[str] = []
) -> Path:
    """
    Run BridgeStan's Makefile on a ``.stan`` file, creating the ``.so``
    used by the StanModel class.

    This function assumes that the path to BridgeStan is valid.
    This can be set with :func:`set_bridgestan_path`.

    :param stan_file: A path to a Stan model file.
    :param stanc_args: A list of arguments to pass to stanc3.
        For example, ``["--O1"]`` will enable compiler optimization level 1.
    :param make_args: A list of additional arguments to pass to Make.
        For example, ``["STAN_THREADS=True"]`` will enable
        threading for the compiled model. If the same flags are defined
        in ``make/local``, the versions passed here will take precedent.
    :raises FileNotFoundError or PermissionError: If `stan_file` does not exist
        or is not readable.
    :raises ValueError: If BridgeStan cannot be located.
    :raises RuntimeError: If compilation fails.
    """
    verify_bridgestan_path(get_bridgestan_path())

    file_path = Path(stan_file).resolve()
    validate_readable(str(file_path))
    if file_path.suffix != ".stan":
        raise ValueError(f"File '{stan_file}' does not end in .stan")

    stanc_args
    output = generate_so_name(file_path)
    cmd = (
        [MAKE]
        + make_args
        + ["STANCFLAGS=" + " ".join(["--include-paths=."] + stanc_args)]
        + [str(output)]
    )
    proc = subprocess.run(
        cmd, cwd=get_bridgestan_path(), capture_output=True, text=True, check=False
    )

    if proc.returncode:
        error = (
            f"Command {' '.join(cmd)} failed with code {proc.returncode}.\n"
            f"stdout:\n{proc.stdout}\nstderr:\n{proc.stderr}"
        )

        raise RuntimeError(error)
    return output



import ctypes
import warnings
from typing import List, Optional, Tuple

import numpy as np
import numpy.typing as npt
from numpy.ctypeslib import ndpointer

from .__version import __version_info__
from .compile import compile_model
from .util import validate_readable

FloatArray = npt.NDArray[np.float64]
double_array = ndpointer(dtype=ctypes.c_double, flags=("C_CONTIGUOUS"))
star_star_char = ctypes.POINTER(ctypes.c_char_p)
c_print_callback = ctypes.CFUNCTYPE(None, ctypes.POINTER(ctypes.c_char), ctypes.c_int)


@c_print_callback
def _print_callback(s, n):
    print(ctypes.string_at(s, n).decode("utf-8"), end="")


class StanModel:
    """
    A StanModel instance encapsulates a Stan model instantiated with data
    and provides methods to access parameter names, transforms, log
    densities, gradients, and Hessians.
    """

    def __init__(
        self,
        model_lib: str,
        model_data: Optional[str] = None,
        *,
        seed: int = 1234,
        capture_stan_prints: bool = True,
    ) -> None:
        """
        Construct a StanModel object for a compiled Stan model and data given
        constructor arguments.

        :param model_lib: A system path to compiled shared object.
        :param model_data: Either a JSON string literal, a
            system path to a data file in JSON format ending in ``.json``,
            or the empty string.
        :param seed: A pseudo random number generator seed, used for RNG functions
            in the ``transformed data`` block.
        :param capture_stan_prints: If ``True``, capture all ``print`` statements
            from the Stan model and print them from Python. This has no effect if
            the model does not contain any ``print`` statements, but may have
            a performance impact if it does. If ``False``, ``print`` statements
            from the Stan model will be sent to ``cout`` and will not be seen in
            Jupyter or capturable with :func:`contextlib.redirect_stdout`.

            **Note:** If this is set for a model, any other models instantiated
            from the *same shared library* will also have the callback set, even
            if they were created *before* this model.
        :raises FileNotFoundError or PermissionError: If ``model_lib`` is not readable or
            ``model_data`` is specified and not a path to a readable file.
        :raises RuntimeError: If there is an error instantiating the
            model from C++.
        """
        validate_readable(model_lib)
        if  model_data is not None and model_data.endswith(".json"):
            validate_readable(model_data)
            with open(model_data, "r") as f:
                model_data = f.read()

        self.lib_path = model_lib
        self.stanlib = ctypes.CDLL(self.lib_path)

        self.data = model_data or ""
        self.seed = seed

        self._construct = self.stanlib.bs_model_construct
        self._construct.restype = ctypes.c_void_p
        self._construct.argtypes = [
            ctypes.c_char_p,
            ctypes.c_uint,
            star_star_char,
        ]

        self._free_error = self.stanlib.bs_free_error_msg
        self._free_error.restype = None
        self._free_error.argtypes = [ctypes.c_char_p]

        self._set_print_callback = self.stanlib.bs_set_print_callback
        self._set_print_callback.restype = None
        self._set_print_callback.argtypes = [c_print_callback, star_star_char]
        if capture_stan_prints:
            self._set_print_callback(_print_callback, None)

        err = ctypes.pointer(ctypes.c_char_p())
        self.model = self._construct(str.encode(self.data), self.seed, err)

        if not self.model:
            raise self._handle_error(err.contents, "bs_model_construct")

        if self.model_version() != __version_info__:
            warnings.warn(
                "The version of the compiled model does not match the version of the "
                "Python package. Consider recompiling the model.",
                RuntimeWarning,
            )

        self._name = self.stanlib.bs_name
        self._name.restype = ctypes.c_char_p
        self._name.argtypes = [ctypes.c_void_p]

        self._model_info = self.stanlib.bs_model_info
        self._model_info.restype = ctypes.c_char_p
        self._model_info.argtypes = [ctypes.c_void_p]

        self._param_num = self.stanlib.bs_param_num
        self._param_num.restype = ctypes.c_int
        self._param_num.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]

        self._param_unc_num = self.stanlib.bs_param_unc_num
        self._param_unc_num.restype = ctypes.c_int
        self._param_unc_num.argtypes = [ctypes.c_void_p]

        self._param_names = self.stanlib.bs_param_names
        self._param_names.restype = ctypes.c_char_p
        self._param_names.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            ctypes.c_int,
        ]

        self._param_unc_names = self.stanlib.bs_param_unc_names
        self._param_unc_names.restype = ctypes.c_char_p
        self._param_unc_names.argtypes = [ctypes.c_void_p]

        self._param_constrain = self.stanlib.bs_param_constrain
        self._param_constrain.restype = ctypes.c_int
        self._param_constrain.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            ctypes.c_int,
            double_array,
            double_array,
            ctypes.c_void_p,
            star_star_char,
        ]

        self._param_unconstrain = self.stanlib.bs_param_unconstrain
        self._param_unconstrain.restype = ctypes.c_int
        self._param_unconstrain.argtypes = [
            ctypes.c_void_p,
            double_array,
            double_array,
            star_star_char,
        ]

        self._param_unconstrain_json = self.stanlib.bs_param_unconstrain_json
        self._param_unconstrain_json.restype = ctypes.c_int
        self._param_unconstrain_json.argtypes = [
            ctypes.c_void_p,
            ctypes.c_char_p,
            double_array,
            star_star_char,
        ]

        self._log_density = self.stanlib.bs_log_density
        self._log_density.restype = ctypes.c_int
        self._log_density.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            ctypes.c_int,
            double_array,
            ctypes.POINTER(ctypes.c_double),
            star_star_char,
        ]

        self._log_density_gradient = self.stanlib.bs_log_density_gradient
        self._log_density_gradient.restype = ctypes.c_int
        self._log_density_gradient.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            ctypes.c_int,
            double_array,
            ctypes.POINTER(ctypes.c_double),
            double_array,
            star_star_char,
        ]

        self._log_density_hessian = self.stanlib.bs_log_density_hessian
        self._log_density_hessian.restype = ctypes.c_int
        self._log_density_hessian.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            ctypes.c_int,
            double_array,
            ctypes.POINTER(ctypes.c_double),
            double_array,
            double_array,
            star_star_char,
        ]

        self._destruct = self.stanlib.bs_model_destruct
        self._destruct.restype = None
        self._destruct.argtypes = [ctypes.c_void_p]

    @classmethod
    def from_stan_file(
        cls,
        stan_file: str,
        model_data: Optional[str] = None,
        *,
        stanc_args: List[str] = [],
        make_args: List[str] = [],
        seed: int = 1234,
        capture_stan_prints: bool = True,
    ):
        """
        Construct a StanModel instance from a ``.stan`` file, compiling if necessary.

        This is equivalent to calling :func:`bridgestan.compile_model`` and then the
        constructor of this class.

        :param stan_file: A path to a Stan model file.
        :param model_data: A path to data in JSON format.
        :param stanc_args: A list of arguments to pass to stanc3.
            For example, ``["--O1"]`` will enable compiler optimization level 1.
        :param make_args: A list of additional arguments to pass to Make.
            For example, ``["STAN_THREADS=True"]`` will enable
            threading for the compiled model. If the same flags are defined
            in ``make/local``, the versions passed here will take precedent.
        :param seed: A pseudo random number generator seed, used for RNG functions
            in the ``transformed data`` block.
        :param capture_stan_prints: If ``True``, capture all ``print`` statements
            from the Stan model and print them from Python. This has no effect if
            the model does not contain any ``print`` statements, but may have
            a performance impact if it does. If ``False``, ``print`` statements
            from the Stan model will be sent to ``cout`` and will not be seen in
            Jupyter or capturable with ``contextlib.redirect_stdout``.
        :raises FileNotFoundError or PermissionError: If ``stan_file`` does not exist
            or is not readable.
        :raises ValueError: If BridgeStan cannot be located.
        :raises RuntimeError: If compilation fails.
        """
        result = compile_model(stan_file, stanc_args=stanc_args, make_args=make_args)
        return cls(
            str(result), model_data, seed=seed, capture_stan_prints=capture_stan_prints
        )

    def __del__(self) -> None:
        """
        Destroy the Stan model and free memory.
        """
        if hasattr(self, "model") and hasattr(self, "_destruct"):
            self._destruct(self.model)

    def __repr__(self) -> str:
        data = f"{self.data!r}, " if self.data else ""
        return f"StanModel({self.lib_path!r}, {data}, seed={self.seed})"

    def name(self) -> str:
        """
        Return the name of the Stan model.

        :return: The name of Stan model.
        """
        return self._name(self.model).decode("utf-8")

    def model_info(self) -> str:
        """
        Return compilation information about the model. For example,
        this includes the current Stan version and important
        compiler settings.

        :return: Information about the compiled Stan model.
        """
        return self._model_info(self.model).decode("utf-8")

    def model_version(self) -> Tuple[int, int, int]:
        """
        Return the BridgeStan version of the compiled model.
        """
        return (
            ctypes.c_int.in_dll(self.stanlib, "bs_major_version").value,
            ctypes.c_int.in_dll(self.stanlib, "bs_minor_version").value,
            ctypes.c_int.in_dll(self.stanlib, "bs_patch_version").value,
        )

    def param_num(self, *, include_tp: bool = False, include_gq: bool = False) -> int:
        """
        Return the number of parameters, including transformed
        parameters and/or generated quantities as indicated.

        :param include_tp: ``True`` to include the transformed parameters.
        :param include_gq: ``True`` to include the generated quantities.
        :return: The number of parameters.
        """
        return self._param_num(self.model, int(include_tp), int(include_gq))

    def param_unc_num(self) -> int:
        """
        Return the number of unconstrained parameters.

        :return: The number of unconstrained parameters.
        """
        return self._param_unc_num(self.model)

    def param_names(
        self, *, include_tp: bool = False, include_gq: bool = False
    ) -> List[str]:
        """
        Return the indexed names of the parameters, including transformed
        parameters and/or generated quantities as indicated.  For
        containers, indexes are separated by periods (`.`).
        For example, the scalar ``a`` has
        indexed name ``a``, the vector entry ``a[1]`` has indexed name ``a.1``
        and the matrix entry ``a[2, 3]`` has indexed name ``a.2.3``.
        Parameter order of the output is column major and more
        generally last-index major for containers.

        :param include_tp: ``True`` to include transformed parameters.
        :param include_gq: ``True`` to include generated quantities.
        :return: The indexed names of the parameters.
        """
        return (
            self._param_names(self.model, int(include_tp), int(include_gq))
            .decode("utf-8")
            .split(",")
        )

    def param_unc_names(self) -> List[str]:
        """
        Return the indexed names of the unconstrained parameters.
        For example, a scalar unconstrained parameter ``b`` has indexed
        name ``b`` and a vector entry ``b[3]`` has indexed name ``b.3``.

        :return: The indexed names of the unconstrained parameters.
        """
        return self._param_unc_names(self.model).decode("utf-8").split(",")

    def param_constrain(
        self,
        theta_unc: FloatArray,
        *,
        include_tp: bool = False,
        include_gq: bool = False,
        out: Optional[FloatArray] = None,
        rng: Optional["StanRNG"] = None,
    ) -> FloatArray:
        """
        Return the constrained parameters derived from the specified
        unconstrained parameters as an array, optionally including the
        transformed parameters and/or generated quantitities as specified.
        Including generated quantities uses the PRNG and may update its state.
        Setting ``out`` avoids allocation of a new array for the return value.

        :param theta_unc: Unconstrained parameter array.
        :param include_tp: ``True`` to include transformed parameters.
        :param include_gq: ``True`` to include generated quantities.
        :param out: A location into which the result is stored.  If
            provided, it must have shape ``(D, )``, where ``D`` is the number of
            constrained parameters.  If not provided or ``None``, a freshly
            allocated array is returned.
        :param rng: A ``StanRNG`` object to use for generating random
            numbers, see :meth:`~StanModel.new_rng``. Must be specified
            if ``include_gq`` is ``True``.
        :return: The constrained parameter array.
        :raises ValueError: If ``out`` is specified and is not the same
            shape as the return.
        :raises ValueError: If ``rng`` is ``None`` and ``include_gq`` is ``True``.
        :raises RuntimeError: If the C++ Stan model throws an exception.
        """
        if rng is None:
            if include_gq:
                raise ValueError(
                    "Error: must specify rng when including generated quantities"
                )
            rng_ptr = None
        else:
            rng_ptr = rng.ptr

        dims = self.param_num(include_tp=include_tp, include_gq=include_gq)
        if out is None:
            out = np.zeros(dims)
        elif out.size != dims:
            raise ValueError(
                "Error: out must be same size as number of constrained parameters"
            )

        err = ctypes.pointer(ctypes.c_char_p())

        rc = self._param_constrain(
            self.model,
            int(include_tp),
            int(include_gq),
            theta_unc,
            out,
            rng_ptr,
            err,
        )

        if rc:
            raise self._handle_error(err.contents, "param_constrain")
        return out

    def new_rng(self, seed) -> "StanRNG":
        """
        Return a new PRNG for use in :meth:`~StanModel.param_constrain``.

        :param seed: A seed for the PRNG.
        :return: A new PRNG wrapper.
        """
        return StanRNG(self.stanlib, seed)

    def param_unconstrain(
        self, theta: FloatArray, *, out: Optional[FloatArray] = None
    ) -> FloatArray:
        """
        Return the unconstrained parameters derived from the specified
        constrained parameters.  Setting ``out`` avoids allocation of a
        new array for the return value.

        :param theta: Constrained parameter array.
        :param out: A location into which the result is stored.  If
            provided, it must have shape ``(D, )``, where ``D`` is the number of
            unconstrained parameters.  If not provided or ``None``, a freshly
            allocated array is returned.
        :raises ValueError: If ``out`` is specified and is not the same
            shape as the return.
        :raises RuntimeError: If the C++ Stan model throws an exception.
        """
        dims = self.param_unc_num()
        if out is None:
            out = np.zeros(shape=dims)
        elif out.size != dims:
            raise ValueError(
                f"out size = {out.size} != unconstrained params size = {dims}"
            )
        err = ctypes.pointer(ctypes.c_char_p())
        rc = self._param_unconstrain(self.model, theta, out, err)
        if rc:
            raise self._handle_error(err.contents, "param_unconstrain")
        return out

    def param_unconstrain_json(
        self, theta_json: str, *, out: Optional[FloatArray] = None
    ) -> FloatArray:
        """
        Return an array of the unconstrained parameters derived from the
        specified JSON formatted data.  See the *CmdStan Reference
        Manual* for the schema definition used.

        :param theta_json: The JSON encoded constrained parameters.
        :param out: A location into which the result is stored.  If
            provided, it must have shape ``(D, )``, where ``D`` is the number of
            unconstrained parameters.  If not provided or ``None``, a freshly
            allocated array is returned.
        :return: The unconstrained parameter array.
        :raises ValueError: If ``out`` is specified and is not the same
            shape as the return value.
        :raises RuntimeError: If the C++ Stan model throws an exception.
        """
        dims = self.param_unc_num()
        if out is None:
            out = np.zeros(shape=dims)
        elif out.size != dims:
            raise ValueError(
                f"out size = {out.size} != unconstrained params size = {dims}"
            )
        chars = theta_json.encode("UTF-8")
        err = ctypes.pointer(ctypes.c_char_p())
        rc = self._param_unconstrain_json(self.model, chars, out, err)
        if rc:
            raise self._handle_error(err.contents, "param_unconstrain_json")
        return out

    def log_density(
        self,
        theta_unc: FloatArray,
        *,
        propto: bool = True,
        jacobian: bool = True,
    ) -> float:
        """
        Return the log density of the specified unconstrained
        parameters, dropping constant terms that do not depend on the
        parameters if ``propto`` is ``True`` and including change of
        variables terms for constrained parameters if ``jacobian`` is ``True``.

        :param theta_unc: Unconstrained parameter array.
        :param propto: ``True`` if constant terms should be dropped from the log density.
        :param jacobian: ``True`` if change-of-variables terms for
            constrained parameters should be included in the log density.
        :return: The log density.
        :raises RuntimeError: If the C++ Stan model throws an exception.
        """
        lp = ctypes.pointer(ctypes.c_double())
        err = ctypes.pointer(ctypes.c_char_p())
        rc = self._log_density(
            self.model, int(propto), int(jacobian), theta_unc, lp, err
        )
        if rc:
            raise self._handle_error(err.contents, "log_density")
        return lp.contents.value

    def log_density_gradient(
        self,
        theta_unc: FloatArray,
        *,
        propto: bool = True,
        jacobian: bool = True,
        out: Optional[FloatArray] = None,
    ) -> Tuple[float, FloatArray]:
        """
        Return a tuple of the log density and gradient of the specified
        unconstrained parameters, dropping constant terms that do not depend
        on the parameters if ``propto`` is ``True`` and including change of
        variables terms for constrained parameters if ``jacobian``
        is ``True``.

        :param theta_unc: Unconstrained parameter array.
        :param propto: ``True`` if constant terms should be dropped from the log density.
        :param jacobian: ``True`` if change-of-variables terms for
            constrained parameters should be included in the log density.
        :param out: A location into which the gradient is stored.  If
            provided, it must have shape `(D, )` where ``D`` is the number
            of parameters.  If not provided, a freshly allocated array
            is returned.
        :return: A tuple consisting of log density and gradient.
        :raises ValueError: If ``out`` is specified and is not the same
            shape as the gradient.
        :raises RuntimeError: If the C++ Stan model throws an exception.
        """
        dims = self.param_unc_num()
        if out is None:
            out = np.zeros(shape=dims)
        elif out.size != dims:
            raise ValueError(f"out size = {out.size} != params size = {dims}")
        lp = ctypes.pointer(ctypes.c_double())
        err = ctypes.pointer(ctypes.c_char_p())
        rc = self._log_density_gradient(
            self.model, int(propto), int(jacobian), theta_unc, lp, out, err
        )
        if rc:
            raise self._handle_error(err.contents, "log_density_gradient")
        return lp.contents.value, out

    def log_density_hessian(
        self,
        theta_unc: FloatArray,
        *,
        propto: bool = True,
        jacobian: bool = True,
        out_grad: Optional[FloatArray] = None,
        out_hess: Optional[FloatArray] = None,
    ) -> Tuple[float, FloatArray, FloatArray]:
        """
        Return a tuple of the log density, gradient, and Hessian of the
        specified unconstrained parameters, dropping constant terms that do
        not depend on the parameters if ``propto`` is ``True`` and including
        change of variables terms for constrained parameters if
        ``jacobian`` is ``True``.

        :param theta_unc: Unconstrained parameter array.
        :param propto: ``True`` if constant terms should be dropped from the log density.
        :param jacobian: ``True`` if change-of-variables terms for
            constrained parameters should be included in the log density.
        :param out_grad: A location into which the gradient is stored.  If
            provided, it must have shape `(D, )` where ``D`` is the number
            of parameters.  If not provided, a freshly allocated array
            is returned.
        :param out_hess: A location into which the Hessian is stored. If
            provided, it must have shape `(D, D)`, where ``D`` is the
            number of parameters.  If not provided, a freshly allocated
            array is returned.
        :return: A tuple consisting of the log density, gradient, and Hessian.
        :raises ValueError: If ``out_grad`` is specified and is not the
            same shape as the gradient or if ``out_hess`` is specified and it
            is not the same shape as the Hessian.
        :raises RuntimeError: If the C++ Stan model throws an exception.
        """
        dims = self.param_unc_num()
        if out_grad is None:
            out_grad = np.zeros(shape=dims)
        elif out_grad.shape != (dims,):
            raise ValueError(f"out_grad size = {out_grad.size} != params size = {dims}")
        hess_size = dims * dims
        if out_hess is None:
            out_hess = np.zeros(shape=hess_size)
        elif out_hess.shape != (dims, dims):
            raise ValueError(
                f"out_hess size = {out_hess.size} != params size^2 = {hess_size}"
            )
        lp = ctypes.pointer(ctypes.c_double())
        err = ctypes.pointer(ctypes.c_char_p())
        rc = self._log_density_hessian(
            self.model,
            int(propto),
            int(jacobian),
            theta_unc,
            lp,
            out_grad,
            out_hess,
            err,
        )
        if rc:
            raise self._handle_error(err.contents, "log_density_hessian")
        out_hess = out_hess.reshape(dims, dims)
        return lp.contents.value, out_grad, out_hess

    def _handle_error(self, err: ctypes.c_char_p, method: str) -> Exception:
        """
        Creates an exception based on a string from C++,
        frees the string, and returns the exception.

        :param err: A C string containing an error message, or nullptr.
        :param method: The name of the method that threw the error.
        :return: An exception based on the.
        """
        if err:
            string = ctypes.string_at(err).decode("utf-8")
            self._free_error(err)
            return RuntimeError(string)
        else:
            return RuntimeError(f"Unknown error in {method}. ")


class StanRNG:
    def __init__(self, lib: ctypes.CDLL, seed: int) -> None:
        """
        Construct a Stan random number generator.
        This should not be called directly. Instead, use
        :meth:`StanModel.new_rng`.
        """
        self.stanlib = lib

        construct = self.stanlib.bs_rng_construct
        construct.restype = ctypes.c_void_p
        construct.argtypes = [ctypes.c_uint, star_star_char]
        self.ptr = construct(seed, None)

        if not self.ptr:
            raise RuntimeError("Failed to construct RNG.")

        self._destruct = self.stanlib.bs_rng_destruct
        self._destruct.restype = None
        self._destruct.argtypes = [ctypes.c_void_p]

    def __del__(self) -> None:
        """
        Destroy the Stan model and free memory.
        """
        if hasattr(self, "ptr") and hasattr(self, "_destruct"):
            self._destruct(self.ptr)



import os
from unittest import mock

import bridgestan as bs


@mock.patch.dict(os.environ, {"BRIDGESTAN": ""})
def test_download_bridgestan():
    bs.compile.verify_bridgestan_path(bs.compile.get_bridgestan_path())



import os


def validate_readable(f: str) -> None:
    """
    Raise an error if the specified file is not readable.

    :param f: The file name.
    :raises FileNotFoundError: If the file is not found.
    :raises PermissionError: If the file is not readable.
    """
    if not os.path.isfile(f):
        raise FileNotFoundError(f"File '{f}' does not exist")
    if not os.access(f, os.R_OK):
        raise PermissionError(f"File '{f}' is not readable")



from ast import literal_eval
from pathlib import Path

from setuptools import setup

VERSIONFILE = Path(__file__).parent / "bridgestan" / "__version.py"
with open(VERSIONFILE, "rt") as f:
    version = literal_eval(f.readline().split("= ")[1])

# remainder of setup in setup.cfg
setup(version=version)



from pathlib import Path

import numpy as np
import pytest

import bridgestan as bs

STAN_FOLDER = Path(__file__).parent.parent.parent / "test_models"


def test_constructor():

    # implicit destructor tests in success and fail cases

    # test empty data
    std_so = str(STAN_FOLDER / "stdnormal" / "stdnormal_model.so")
    b1 = bs.StanModel(std_so)
    np.testing.assert_allclose(bool(b1), True)

    # test load data
    bernoulli_so = str(STAN_FOLDER / "bernoulli" / "bernoulli_model.so")
    bernoulli_data = str(STAN_FOLDER / "bernoulli" / "bernoulli.data.json")
    b2 = bs.StanModel(bernoulli_so, bernoulli_data)
    np.testing.assert_allclose(bool(b2), True)

    bernoulli_data_string = (
        STAN_FOLDER / "bernoulli" / "bernoulli.data.json"
    ).read_text()
    b3 = bs.StanModel(bernoulli_so, bernoulli_data_string)
    np.testing.assert_allclose(bool(b3), True)

    # test missing so file
    with pytest.raises(FileNotFoundError):
        bs.StanModel("nope, not going to find it")

    # test missing data file
    with pytest.raises(FileNotFoundError):
        bs.StanModel(bernoulli_so, "nope, not going to find it.json")

    # test data load exception
    throw_data_so = str(STAN_FOLDER / "throw_data" / "throw_data_model.so")
    with pytest.raises(RuntimeError, match="find this text: datafails"):
        b4 = bs.StanModel(throw_data_so)


def test_name():
    std_so = str(STAN_FOLDER / "stdnormal" / "stdnormal_model.so")
    b = bs.StanModel(std_so)
    np.testing.assert_equal("stdnormal_model", b.name())


def test_model_info():
    std_so = str(STAN_FOLDER / "stdnormal" / "stdnormal_model.so")
    b = bs.StanModel(std_so)
    assert "STAN_OPENCL" in b.model_info()
    assert "BridgeStan version: 1." in b.model_info()


def test_param_num():
    full_so = str(STAN_FOLDER / "full" / "full_model.so")
    b = bs.StanModel(full_so)
    np.testing.assert_equal(1, b.param_num())
    np.testing.assert_equal(1, b.param_num(include_tp=False))
    np.testing.assert_equal(1, b.param_num(include_gq=False))
    np.testing.assert_equal(1, b.param_num(include_tp=False, include_gq=False))
    np.testing.assert_equal(3, b.param_num(include_gq=True))
    np.testing.assert_equal(3, b.param_num(include_tp=False, include_gq=True))
    np.testing.assert_equal(2, b.param_num(include_tp=True))
    np.testing.assert_equal(2, b.param_num(include_tp=True, include_gq=False))
    np.testing.assert_equal(4, b.param_num(include_tp=True, include_gq=True))


def test_param_unc_num():
    simplex_so = str(STAN_FOLDER / "simplex" / "simplex_model.so")
    b = bs.StanModel(simplex_so)
    np.testing.assert_equal(5, b.param_num())
    np.testing.assert_equal(4, b.param_unc_num())


def test_param_names():
    matrix_so = str(STAN_FOLDER / "matrix" / "matrix_model.so")
    b = bs.StanModel(matrix_so)
    np.testing.assert_array_equal(
        ["A.1.1", "A.2.1", "A.3.1", "A.1.2", "A.2.2", "A.3.2"], b.param_names()
    )
    np.testing.assert_array_equal(
        ["A.1.1", "A.2.1", "A.3.1", "A.1.2", "A.2.2", "A.3.2"],
        b.param_names(include_tp=False),
    )
    np.testing.assert_array_equal(
        ["A.1.1", "A.2.1", "A.3.1", "A.1.2", "A.2.2", "A.3.2"],
        b.param_names(include_gq=False),
    )
    np.testing.assert_array_equal(
        ["A.1.1", "A.2.1", "A.3.1", "A.1.2", "A.2.2", "A.3.2"],
        b.param_names(include_tp=False, include_gq=False),
    )
    np.testing.assert_array_equal(
        [
            "A.1.1",
            "A.2.1",
            "A.3.1",
            "A.1.2",
            "A.2.2",
            "A.3.2",
            "B.1.1",
            "B.2.1",
            "B.3.1",
            "B.1.2",
            "B.2.2",
            "B.3.2",
        ],
        b.param_names(include_tp=True),
    )
    np.testing.assert_array_equal(
        [
            "A.1.1",
            "A.2.1",
            "A.3.1",
            "A.1.2",
            "A.2.2",
            "A.3.2",
            "B.1.1",
            "B.2.1",
            "B.3.1",
            "B.1.2",
            "B.2.2",
            "B.3.2",
        ],
        b.param_names(include_tp=True, include_gq=False),
    )
    np.testing.assert_array_equal(
        ["A.1.1", "A.2.1", "A.3.1", "A.1.2", "A.2.2", "A.3.2", "c"],
        b.param_names(include_gq=True),
    )
    np.testing.assert_array_equal(
        ["A.1.1", "A.2.1", "A.3.1", "A.1.2", "A.2.2", "A.3.2", "c"],
        b.param_names(include_tp=False, include_gq=True),
    )
    np.testing.assert_array_equal(
        [
            "A.1.1",
            "A.2.1",
            "A.3.1",
            "A.1.2",
            "A.2.2",
            "A.3.2",
            "B.1.1",
            "B.2.1",
            "B.3.1",
            "B.1.2",
            "B.2.2",
            "B.3.2",
            "c",
        ],
        b.param_names(include_tp=True, include_gq=True),
    )


def test_param_unc_names():
    matrix_so = str(STAN_FOLDER / "matrix" / "matrix_model.so")
    b1 = bs.StanModel(matrix_so)
    np.testing.assert_array_equal(
        ["A.1.1", "A.2.1", "A.3.1", "A.1.2", "A.2.2", "A.3.2"], b1.param_unc_names()
    )

    simplex_so = str(STAN_FOLDER / "simplex" / "simplex_model.so")
    b2 = bs.StanModel(simplex_so)
    np.testing.assert_array_equal(
        ["theta.1", "theta.2", "theta.3", "theta.4"], b2.param_unc_names()
    )


def cov_constrain(v, D):
    L = np.zeros([D, D])
    idxL = np.tril_indices(D)
    L[idxL] = v
    idxD = np.diag_indices(D)
    L[idxD] = np.exp(L[idxD])
    return np.matmul(L, L.T)


def test_param_constrain():
    fr_gaussian_so = str(STAN_FOLDER / "fr_gaussian" / "fr_gaussian_model.so")
    fr_gaussian_data = str(STAN_FOLDER / "fr_gaussian" / "fr_gaussian.data.json")
    bridge = bs.StanModel(fr_gaussian_so, fr_gaussian_data)

    D = 4
    unc_size = 10
    a = np.random.normal(size=unc_size)
    B_expected = cov_constrain(a, D)

    b = bridge.param_constrain(a, include_tp=False, include_gq=False)
    B = b.reshape(D, D)
    np.testing.assert_allclose(B_expected, B)

    b = bridge.param_constrain(a, include_gq=False)
    B = b.reshape(D, D)
    np.testing.assert_allclose(B_expected, B)

    b = bridge.param_constrain(a, include_tp=False)
    B = b.reshape(D, D)
    np.testing.assert_allclose(B_expected, B)

    b = bridge.param_constrain(a)
    B = b.reshape(D, D)
    np.testing.assert_allclose(B_expected, B)

    full_so = str(STAN_FOLDER / "full" / "full_model.so")
    bridge2 = bs.StanModel(full_so)
    rng = bridge2.new_rng(seed=1234)

    np.testing.assert_equal(1, bridge2.param_constrain(a).size)
    np.testing.assert_equal(2, bridge2.param_constrain(a, include_tp=True).size)
    np.testing.assert_equal(
        3, bridge2.param_constrain(a, include_gq=True, rng=rng).size
    )
    np.testing.assert_equal(
        4, bridge2.param_constrain(a, include_tp=True, include_gq=True, rng=rng).size
    )

    # reproducibility test
    np.testing.assert_equal(
        bridge2.param_constrain(a, include_gq=True, rng=bridge2.new_rng(seed=4567)),
        bridge2.param_constrain(a, include_gq=True, rng=bridge2.new_rng(seed=4567)),
    )

    # test error if neither seed or rng is provided
    with pytest.raises(ValueError):
        bridge2.param_constrain(a, include_gq=True)

    # out tests, matched and mismatched
    scratch = np.zeros(16)
    b = bridge.param_constrain(a, out=scratch)
    B = b.reshape(D, D)
    np.testing.assert_allclose(B_expected, B)
    scratch_wrong = np.zeros(10)
    with pytest.raises(ValueError):
        bridge.param_constrain(a, out=scratch_wrong)

    # exception handling test in transformed parameters/model (compiled same way)
    throw_tp_so = str(STAN_FOLDER / "throw_tp" / "throw_tp_model.so")
    bridge2 = bs.StanModel(throw_tp_so)

    y = np.array(np.random.uniform(1))
    bridge2.param_constrain(y, include_tp=False)
    with pytest.raises(RuntimeError, match="find this text: tpfails"):
        bridge2.param_constrain(y, include_tp=True)

    throw_gq_so = str(STAN_FOLDER / "throw_gq" / "throw_gq_model.so")
    bridge3 = bs.StanModel(throw_gq_so)
    bridge3.param_constrain(y, include_gq=False)
    with pytest.raises(RuntimeError, match="find this text: gqfails"):
        bridge3.param_constrain(y, include_gq=True, rng=bridge3.new_rng(seed=1))


def test_param_unconstrain():
    fr_gaussian_so = str(STAN_FOLDER / "fr_gaussian" / "fr_gaussian_model.so")
    fr_gaussian_data = str(STAN_FOLDER / "fr_gaussian" / "fr_gaussian.data.json")
    bridge = bs.StanModel(fr_gaussian_so, fr_gaussian_data)

    unc_size = 10
    a = np.random.normal(size=unc_size)
    b = bridge.param_constrain(a)
    c = bridge.param_unconstrain(b)
    np.testing.assert_allclose(a, c)

    scratch = np.zeros(10)
    c2 = bridge.param_unconstrain(b, out=scratch)
    np.testing.assert_allclose(a, c2)
    scratch_wrong = np.zeros(16)
    with pytest.raises(ValueError):
        bridge.param_unconstrain(b, out=scratch_wrong)


def test_param_unconstrain_json():
    gaussian_so = str(STAN_FOLDER / "gaussian" / "gaussian_model.so")
    gaussian_data = str(STAN_FOLDER / "gaussian" / "gaussian.data.json")
    bridge = bs.StanModel(gaussian_so, gaussian_data)

    # theta = np.array([0.2, 1.9])
    theta_unc = np.array([0.2, np.log(1.9)])
    theta_json = '{"mu": 0.2, "sigma": 1.9}'
    theta_unc_j_test = bridge.param_unconstrain_json(theta_json)
    np.testing.assert_allclose(theta_unc, theta_unc_j_test)

    scratch = np.zeros(2)
    theta_unc_j_test2 = bridge.param_unconstrain_json(theta_json, out=scratch)
    np.testing.assert_allclose(theta_unc, theta_unc_j_test2)

    scratch_bad = np.zeros(10)
    with pytest.raises(ValueError):
        bridge.param_unconstrain_json(theta_json, out=scratch_bad)


def _log_jacobian(p):
    return np.log(p * (1 - p))


def _bernoulli(y, p):
    return np.sum(y * np.log(p) + (1 - y) * np.log(1 - p))


def _bernoulli_jacobian(y, p):
    return _bernoulli(y, p) + _log_jacobian(p)


def test_log_density():
    bernoulli_so = str(STAN_FOLDER / "bernoulli" / "bernoulli_model.so")
    bernoulli_data = str(STAN_FOLDER / "bernoulli" / "bernoulli.data.json")
    bridge = bs.StanModel(bernoulli_so, bernoulli_data)
    y = np.asarray([0, 1, 0, 0, 0, 0, 0, 0, 0, 1])
    for _ in range(2):
        x = np.random.uniform(size=bridge.param_unc_num())
        x_unc = np.log(x / (1 - x))
        lp = bridge.log_density(np.array([x_unc]), propto=False, jacobian=False)
        np.testing.assert_allclose(lp, _bernoulli(y, x))
        lp2 = bridge.log_density(np.array([x_unc]), propto=False, jacobian=True)
        np.testing.assert_allclose(lp2, _bernoulli_jacobian(y, x))
        lp3 = bridge.log_density(np.array([x_unc]), propto=True, jacobian=True)
        np.testing.assert_allclose(lp3, _bernoulli_jacobian(y, x))
        lp4 = bridge.log_density(np.array([x_unc]), propto=True, jacobian=False)
        np.testing.assert_allclose(lp4, _bernoulli(y, x))

    throw_lp_so = str(STAN_FOLDER / "throw_lp" / "throw_lp_model.so")
    bridge2 = bs.StanModel(throw_lp_so)
    y2 = np.array(np.random.uniform(1))

    with pytest.raises(RuntimeError, match="find this text: lpfails"):
        bridge2.log_density(y2)


def test_log_density_gradient():
    def _logp(y_unc):
        y = np.exp(y_unc)
        return -0.5 * y**2

    def _propto_false(y_unc):
        return -0.5 * np.log(2 * np.pi)

    def _jacobian_true(y_unc):
        return y_unc

    def _grad_logp(y_unc):
        y = np.exp(y_unc)
        return -(y**2)

    def _grad_propto_false(y_unc):
        return 0

    def _grad_jacobian_true(y_unc):
        return 1

    jacobian_so = str(STAN_FOLDER / "jacobian" / "jacobian_model.so")
    bridge = bs.StanModel(jacobian_so)

    y = np.abs(np.random.normal(1))
    y_unc = np.log(y)
    y_unc_arr = np.array(y_unc)
    logdensity, grad = bridge.log_density_gradient(
        y_unc_arr, propto=True, jacobian=True
    )
    np.testing.assert_allclose(_logp(y_unc) + _jacobian_true(y_unc), logdensity)
    np.testing.assert_allclose(_grad_logp(y_unc) + _grad_jacobian_true(y_unc), grad[0])
    #
    logdensity, grad = bridge.log_density_gradient(
        y_unc_arr, propto=True, jacobian=False
    )
    np.testing.assert_allclose(_logp(y_unc), logdensity)
    np.testing.assert_allclose(_grad_logp(y_unc), grad[0])
    #
    logdensity, grad = bridge.log_density_gradient(
        y_unc_arr, propto=False, jacobian=True
    )
    np.testing.assert_allclose(
        _logp(y_unc) + _propto_false(y_unc) + _jacobian_true(y_unc), logdensity
    )
    np.testing.assert_allclose(
        _grad_logp(y_unc) + _grad_propto_false(y_unc) + _grad_jacobian_true(y_unc),
        grad[0],
    )
    #
    logdensity, grad = bridge.log_density_gradient(
        y_unc_arr, propto=False, jacobian=False
    )
    np.testing.assert_allclose(_logp(y_unc) + _propto_false(y_unc), logdensity)
    np.testing.assert_allclose(_grad_logp(y_unc) + _grad_propto_false(y_unc), grad[0])

    # test use of scratch
    scratch = np.zeros(bridge.param_unc_num())
    logdensity, grad = bridge.log_density_gradient(
        y_unc_arr, propto=True, jacobian=True, out=scratch
    )
    np.testing.assert_allclose(_logp(y_unc) + _jacobian_true(y_unc), logdensity)
    np.testing.assert_allclose(_grad_logp(y_unc) + _grad_jacobian_true(y_unc), grad[0])
    #
    scratch_bad = np.zeros(bridge.param_unc_num() + 10)
    with pytest.raises(ValueError):
        bridge.log_density_gradient(y_unc, out=scratch_bad)


def test_log_density_hessian():
    def _logp(y_unc):
        y = np.exp(y_unc)
        return -0.5 * y**2

    def _propto_false(y_unc):
        return -0.5 * np.log(2 * np.pi)

    def _jacobian_true(y_unc):
        return y_unc

    def _grad_logp(y_unc):
        y = np.exp(y_unc)
        return -(y**2)

    def _grad_propto_false(y_unc):
        return 0

    def _grad_jacobian_true(y_unc):
        return 1

    def _hess_logp(y_unc):
        y = np.exp(y_unc)
        return -2.0 * y**2

    def _hess_propto_false(y_unc):
        return 0

    def _hess_jacobian_true(y_unc):
        return 0

    jacobian_so = str(STAN_FOLDER / "jacobian" / "jacobian_model.so")
    bridge = bs.StanModel(jacobian_so)

    # test value, gradient, hessian, all combos +/- propto, +/- jacobian
    y = np.abs(np.random.normal(1))
    y_unc = np.log(y)
    y_unc_arr = np.array(y_unc)
    logdensity, grad, hess = bridge.log_density_hessian(
        y_unc_arr, propto=True, jacobian=True
    )
    np.testing.assert_allclose(_logp(y_unc) + _jacobian_true(y_unc), logdensity)
    np.testing.assert_allclose(_grad_logp(y_unc) + _grad_jacobian_true(y_unc), grad[0])
    np.testing.assert_allclose(
        _hess_logp(y_unc) + _hess_jacobian_true(y_unc), hess[0, 0]
    )
    #
    logdensity, grad, hess = bridge.log_density_hessian(
        y_unc_arr, propto=True, jacobian=False
    )
    np.testing.assert_allclose(_logp(y_unc), logdensity)
    np.testing.assert_allclose(_grad_logp(y_unc), grad[0])
    np.testing.assert_allclose(_hess_logp(y_unc), hess[0, 0])
    #
    logdensity, grad, hess = bridge.log_density_hessian(
        y_unc_arr, propto=False, jacobian=True
    )
    np.testing.assert_allclose(
        _logp(y_unc) + _propto_false(y_unc) + _jacobian_true(y_unc), logdensity
    )
    np.testing.assert_allclose(
        _grad_logp(y_unc) + _grad_propto_false(y_unc) + _grad_jacobian_true(y_unc),
        grad[0],
    )
    np.testing.assert_allclose(
        _hess_logp(y_unc) + _hess_propto_false(y_unc) + _hess_jacobian_true(y_unc),
        hess[0, 0],
    )
    #
    logdensity, grad, hess = bridge.log_density_hessian(
        y_unc_arr, propto=False, jacobian=True
    )
    np.testing.assert_allclose(
        _logp(y_unc) + _propto_false(y_unc) + _jacobian_true(y_unc), logdensity
    )
    np.testing.assert_allclose(
        _grad_logp(y_unc) + _grad_propto_false(y_unc) + _grad_jacobian_true(y_unc),
        grad[0],
    )
    np.testing.assert_allclose(
        _hess_logp(y_unc) + _hess_propto_false(y_unc) + _hess_jacobian_true(y_unc),
        hess[0, 0],
    )
    #
    logdensity, grad, hess = bridge.log_density_hessian(
        y_unc_arr, propto=False, jacobian=False
    )
    np.testing.assert_allclose(_logp(y_unc) + _propto_false(y_unc), logdensity)
    np.testing.assert_allclose(_grad_logp(y_unc) + _grad_propto_false(y_unc), grad[0])
    np.testing.assert_allclose(
        _hess_logp(y_unc) + _hess_propto_false(y_unc), hess[0, 0]
    )

    # test use of scratch
    scratch = np.zeros(bridge.param_unc_num())
    logdensity, grad, hess = bridge.log_density_hessian(
        y_unc_arr, propto=True, jacobian=True, out_grad=scratch
    )
    np.testing.assert_allclose(_logp(y_unc) + _jacobian_true(y_unc), logdensity)
    np.testing.assert_allclose(_grad_logp(y_unc) + _grad_jacobian_true(y_unc), grad[0])
    #
    scratch_bad = np.zeros(bridge.param_unc_num() + 10)
    with pytest.raises(ValueError):
        bridge.log_density_hessian(y_unc, out_grad=scratch_bad)

    # test with 5 x 5 Hessian
    simple_so = str(STAN_FOLDER / "simple" / "simple_model.so")
    simple_data = str(STAN_FOLDER / "simple" / "simple.data.json")
    bridge2 = bs.StanModel(simple_so, simple_data)

    D = 5
    y = np.random.uniform(size=D)
    lp, grad, hess = bridge2.log_density_hessian(y)
    np.testing.assert_allclose(-y, grad)
    np.testing.assert_allclose(-np.identity(D), hess)


def test_out_behavior():
    bernoulli_so = str(STAN_FOLDER / "bernoulli" / "bernoulli_model.so")
    bernoulli_data = str(STAN_FOLDER / "bernoulli" / "bernoulli.data.json")
    smb = bs.StanModel(bernoulli_so, bernoulli_data)

    grads = []
    for _ in range(2):
        x = np.random.uniform(size=smb.param_unc_num())
        q = np.log(x / (1 - x))  # unconstrained scale
        _, grad = smb.log_density_gradient(q, propto=1, jacobian=0)
        grads.append(grad)

    # default behavior is fresh array
    assert grads[0] is not grads[1]

    grads = []
    out_grad = np.zeros(shape=smb.param_unc_num())
    for _ in range(2):
        x = np.random.uniform(size=smb.param_unc_num())
        q = np.log(x / (1 - x))  # unconstrained scale
        _, grad = smb.log_density_gradient(q, propto=1, jacobian=0, out=out_grad)
        grads.append(out_grad)

    # out parameter is modified and reference is returned
    assert grads[0] is grads[1]
    assert grads[0] is out_grad
    assert grads[1] is out_grad
    np.testing.assert_allclose(grads[0], grads[1])


# BONUS TESTS


def test_bernoulli():
    def _bernoulli(y, p):
        return np.sum(y * np.log(p) + (1 - y) * np.log(1 - p))

    bernoulli_so = str(STAN_FOLDER / "bernoulli" / "bernoulli_model.so")
    bernoulli_data = str(STAN_FOLDER / "bernoulli" / "bernoulli.data.json")
    smb = bs.StanModel(bernoulli_so, bernoulli_data)
    np.testing.assert_string_equal(smb.name(), "bernoulli_model")
    np.testing.assert_allclose(smb.param_unc_num(), 1)
    np.testing.assert_allclose(smb.param_num(include_tp=False, include_gq=False), 1)
    y = np.asarray([0, 1, 0, 0, 0, 0, 0, 0, 0, 1])
    R = 2
    for _ in range(R):
        x = np.random.uniform(size=smb.param_unc_num())
        q = np.log(x / (1 - x))  # unconstrained scale
        logdensity, grad = smb.log_density_gradient(q, propto=True, jacobian=False)
        np.testing.assert_allclose(logdensity, _bernoulli(y, x))
        constrained_theta = smb.param_constrain(q, include_tp=False, include_gq=False)
        np.testing.assert_allclose(constrained_theta, x)
        np.testing.assert_allclose(smb.param_unconstrain(constrained_theta), q)


def test_multi():
    def _multi(x):
        return -0.5 * np.dot(x, x)

    def _grad_multi(x):
        return -x

    multi_so = str(STAN_FOLDER / "multi" / "multi_model.so")
    multi_data = str(STAN_FOLDER / "multi" / "multi.data.json")

    smm = bs.StanModel(multi_so, multi_data)
    x = np.random.normal(size=smm.param_unc_num())
    logdensity, grad = smm.log_density_gradient(x)
    np.testing.assert_allclose(logdensity, _multi(x))
    np.testing.assert_allclose(grad, _grad_multi(x))


def test_gaussian():

    lib = str(STAN_FOLDER / "gaussian" / "gaussian_model.so")
    data = str(STAN_FOLDER / "gaussian" / "gaussian.data.json")

    model = bs.StanModel(lib, data)

    theta = np.array([0.2, 1.9])
    theta_unc = np.array([0.2, np.log(1.9)])

    theta_test = model.param_constrain(theta_unc, include_tp=0, include_gq=0)
    np.testing.assert_allclose(theta, theta_test)

    theta_unc_test = model.param_unconstrain(theta)
    np.testing.assert_allclose(theta_unc, theta_unc_test)

    theta_json = '{"mu": 0.2, "sigma": 1.9}'
    theta_unc_j_test = model.param_unconstrain_json(theta_json)
    np.testing.assert_allclose(theta_unc, theta_unc_j_test)


def test_fr_gaussian():

    lib = str(STAN_FOLDER / "fr_gaussian" / "fr_gaussian_model.so")
    data = str(STAN_FOLDER / "fr_gaussian" / "fr_gaussian.data.json")
    model = bs.StanModel(lib, data)

    size = 16
    unc_size = 10
    np.testing.assert_allclose(model.param_num(include_tp=True, include_gq=True), size)
    np.testing.assert_allclose(model.param_unc_num(), unc_size)

    D = 4
    a = np.random.normal(size=unc_size)
    b = model.param_constrain(a, include_tp=False, include_gq=False)

    B = b.reshape(D, D)
    B_expected = cov_constrain(a, D)
    np.testing.assert_allclose(B_expected, B)

    c = model.param_unconstrain(b)
    np.testing.assert_allclose(a, c)

    names = model.param_names(include_tp=True, include_gq=True)
    pos = 0
    for j in range(1, 5):
        for i in range(1, 5):
            np.testing.assert_string_equal(names[pos], f"Omega.{i}.{j}")
            pos += 1

    names_unc = model.param_unc_names()
    pos = 0
    for n in range(1, 11):
        np.testing.assert_string_equal(names_unc[pos], f"Omega.{n}")
        pos += 1


def test_stdout_capture():
    import contextlib, io

    theta = 0.1

    m = bs.StanModel(
        str(STAN_FOLDER / "print" / "print_model.so"), capture_stan_prints=False
    )

    with contextlib.redirect_stdout(io.StringIO()) as f:
        print("Hello from Python!")
        m.log_density(np.array([theta]))

    captured = f.getvalue()
    assert captured.splitlines()[0] == "Hello from Python!"
    assert "Stan" not in captured

    m2 = bs.StanModel(str(STAN_FOLDER / "print" / "print_model.so"))

    with contextlib.redirect_stdout(io.StringIO()) as f:
        print("Hello from Python!")
        m2.log_density(np.array([theta]))

    lines = f.getvalue().splitlines()
    assert lines[0] == "Hello from Python!"
    assert lines[1] == "Hi from Stan!"
    assert lines[2] == f"theta = {theta}"

    # test re-entrancy
    import ctypes, threading

    # define a new, sillier, callback which lets us test thread safety
    x = 0

    @ctypes.CFUNCTYPE(None, ctypes.POINTER(ctypes.c_char), ctypes.c_int)
    def callback(s, n):
        nonlocal x
        x += 1

    m2._set_print_callback(callback, None)

    # call it many times from several threads
    def f():
        rng = m2.new_rng(1234)
        for _ in range(25):
            m2.param_constrain(np.array([theta]), include_gq=True, rng=rng)

    threads = [threading.Thread(target=f) for _ in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert x == 500  # 2 calls per print, 10 threads, 25 iterations


@pytest.fixture
def recompile_simple():
    """Recompile simple_model with autodiff hessian enable, then clean-up/restore it after test"""

    stanfile = STAN_FOLDER / "simple" / "simple.stan"
    lib = bs.compile.generate_so_name(stanfile)
    lib.unlink(missing_ok=True)
    res = bs.compile_model(stanfile, make_args=["BRIDGESTAN_AD_HESSIAN=true"])

    yield res

    lib.unlink(missing_ok=True)
    bs.compile_model(stanfile, make_args=["STAN_THREADS=true"])


@pytest.mark.ad_hessian
def test_hessian_autodiff(recompile_simple):
    simple_data = str(STAN_FOLDER / "simple" / "simple.data.json")
    model = bs.StanModel(recompile_simple, simple_data)
    assert "BRIDGESTAN_AD_HESSIAN=true" in model.model_info()
    D = 5
    y = np.random.uniform(size=D)
    lp, grad, hess = model.log_density_hessian(y)
    np.testing.assert_allclose(-y, grad)
    np.testing.assert_allclose(-np.identity(D), hess)


if __name__ == "__main__":
    print("")
    print("TESTING BrigeStan Python API")
    print("------------------------------------------------------------")
    print("test: constructor")
    test_constructor()
    print("test: name")
    test_name()
    print("test: param_num")
    test_param_num()
    print("test: param_unc_num")
    test_param_unc_num()
    print("test: param_names")
    test_param_names()
    print("test: param_unc_names")
    test_param_unc_names()
    print("test: param_constrain")
    test_param_constrain()
    print("test: param_unconstrain")
    test_param_unconstrain()
    print("test: param_unconstrain_json")
    test_param_unconstrain_json()
    print("test: log_density")
    test_log_density()
    print("test: log_density_gradient")
    test_log_density_gradient()
    print("test: log_density_hessian")
    test_log_density_hessian()
    print("test: out behavior")
    test_out_behavior()
    print("test: bernoulli")
    test_bernoulli()
    print("test: multi")
    test_multi()
    print("test: gaussian")
    test_gaussian()
    print("test: fr_gaussian")
    test_fr_gaussian()
    print("------------------------------------------------------------")
    print("If no errors were reported, all tests passed.")



from pathlib import Path

import pytest

import bridgestan as bs

STAN_FOLDER = Path(__file__).parent.parent.parent / "test_models"


def test_compile_good():
    stanfile = STAN_FOLDER / "multi" / "multi.stan"
    lib = bs.compile.generate_so_name(stanfile)
    lib.unlink(missing_ok=True)
    res = bs.compile_model(stanfile, stanc_args=["--O1"])
    assert lib.samefile(res)
    lib.unlink()
    res = bs.compile_model(stanfile, make_args=["STAN_THREADS=true"])
    assert lib.samefile(res)


def test_compile_bad_ext():
    not_stanfile = STAN_FOLDER / "multi" / "multi.data.json"
    with pytest.raises(ValueError, match=r".stan"):
        bs.compile_model(not_stanfile)


def test_compile_nonexistant():
    not_stanfile = STAN_FOLDER / "multi" / "multi-nothere.data.json"
    with pytest.raises(FileNotFoundError):
        bs.compile_model(not_stanfile)


def test_compile_syntax_error():
    stanfile = STAN_FOLDER / "syntax_error" / "syntax_error.stan"
    with pytest.raises(RuntimeError, match=r"Syntax error"):
        bs.compile_model(stanfile)


def test_compile_bad_bridgestan():
    with pytest.raises(ValueError, match=r"does not exist"):
        bs.compile.set_bridgestan_path("dummy")
    with pytest.raises(ValueError, match=r"does not contain file 'Makefile'"):
        bs.compile.set_bridgestan_path(str(STAN_FOLDER))
